#!/usr/bin/perl -w
use strict;
use List::Util qw(first max maxstr min minstr reduce shuffle sum);

my $infile_expr = $ARGV[0];
my $outfile1 = $ARGV[1];

open( in1, "<$infile_expr" );
my $header1 = <in1>;
my @data1 = <in1>;
chomp( @data1 );
close( in1 );

open( out1, ">$outfile1" );
for( my $i = 0; $i < $#data1; $i++ ){
	#my ( $id1, $symbol1, $expr1 ) = split( /\t/, $data1[$i], 3 );
	my ( $id1, $expr1 ) = split( /\t/, $data1[$i], 2 );
	print out1 join( "\t", $id1, $id1, 0 ), "\n"; # "0" means the distance is zero
	print "Processing EntrezGene $id1 ....\n";
	for( my $j = $i + 1; $j <= $#data1; $j++ ) {
		#my ( $id2, $symbol2, $expr2 ) = split( /\t/, $data1[$j], 3 );
		my ( $id2, $expr2 ) = split( /\t/, $data1[$j], 2 );
		
		#my $corr1 = corr( $expr1, $expr2 );
		my ( $corr1, $zscore1 ) = split( /\t/, spearman( join( "\n", get_rank( $expr1 ), get_rank( $expr2 ) ) ) );
		#print out1 join( "\t", $id1, $id2, 1 - abs( $corr1 ) ), "\n";
		print out1 join( "\t", $id1, $id2, ( 1 - $corr1 ) / 2 ), "\n";
	}
}
close( out1 );

sub std {
	my @arr_p=@_;
	
	my $mean=sum(@arr_p)/($#arr_p+1);
	
	my $var=0;
	for(my $si=0;$si<=$#arr_p;$si++) {
		my $diff=$arr_p[$si]-$mean;
		$var=$var+$diff*$diff;
	}
	$var=$var/$#arr_p;
	
	return sqrt($var);
}

sub corr {
	my ( $vector1, $vector2 ) = @_;
	
	my @data1 = split(/\t/,$vector1);
	my @data2 = split(/\t/,$vector2);
	
	if($#data1 ne $#data2) {
		print "Error in funcntion (corr): vector sizes do not match. Exiting.\n";
		exit;
	}
	
	my $mean1 = sum(@data1)/($#data1+1);
	my $mean2 = sum(@data2)/($#data2+1);
	my $std1 = std(@data1);
	my $std2 = std(@data2);
	
	if(($mean1 == 0 && $std1==0) || ($mean2==0 && $std2==0)) {
		return "0"; #пȫ0Уֱ
	}
	
	my $corr = 0;
	for( my $i=0; $i <= $#data1; $i++ ) {
		$corr = $corr + ( $data1[$i] - $mean1 ) * ( $data2[$i] - $mean2 );
	}
	
	$corr = $corr / ( $#data1 + 1 ) / $std1 / $std2;
	return $corr;
}

sub spearman {
	my ( $rank_lst1, $rank_lst2 ) = split( /\n/, $_[0] );
	my @rank1 = split( /\t/, $rank_lst1 );
	my @rank2 = split( /\t/, $rank_lst2 );
	
	if( $#rank1 != $#rank2 ) {
		print "The length of rank lists are not matched. exit from spearman().\n";
		exit;
	}
	
	my $n_rank = $#rank1 + 1;
	my $p_corr = 0;
	my $p_corr_zscore = 0;
	for( my $i = 0; $i <= $#rank1; $i++ ) {
		my $delta = $rank1[$i] - $rank2[$i];
		$p_corr = $p_corr + $delta * $delta;
	}
	
	$p_corr = 1 - ( 6 * $p_corr / $n_rank / ( $n_rank * $n_rank - 1 ) );
	
	if( $p_corr == 1 ) {
		$p_corr_zscore = "Inf";
	}
	else {
		if( $p_corr == -1 ) {
			$p_corr_zscore = "-Inf";
		}
		else {
			$p_corr_zscore = sqrt( ( $n_rank - 3 ) / 1.06 ) * 0.5 * log( ( 1 + $p_corr ) / ( 1 - $p_corr ) );
		}
	}
	
	my $short_p_corr = sprintf( "%.3f", $p_corr );
	my $short_p_corr_zscore = sprintf( "%.3f", $p_corr_zscore );
	
	#return join( "\t", $short_p_corr, $short_p_corr_zscore );
	return join( "\t", $p_corr, $p_corr_zscore );
}

sub get_rank {
	my $data_lst1 = $_[0];
	my @data_exp1 = split( /\t/, $data_lst1 );
	my @exp_lst1 = ();
		
	my %exp_rank1 = ();
	
	for( my $i = 0; $i <= $#data_exp1; $i++ ) {
		#my ( $rank1, $tot1, $exp1 ) = split( /:/, $data_exp1[$i] );
		#$exp_rank1{ $exp1 } = $rank1 / $tot1;
		#push( @exp_lst1, $exp1 );
		$exp_rank1{ $i + 1 } = $data_exp1[$i];
		push( @exp_lst1, $i + 1 );
	}
	
	my @sort_exp_rank1 = sort { $exp_rank1{ $a } <=> $exp_rank1{ $b } } keys %exp_rank1;
	my $curr_data1 = $exp_rank1{ $sort_exp_rank1[0] };
	my $accu_rank1 = 1;
	my $curr_rank1 = 1;
	my @store_exp1 = ();
	my $tie_rank1 = 0;
	push( @store_exp1, shift( @sort_exp_rank1 ) );
	
	while( my $next_exp1 = shift( @sort_exp_rank1 ) ) {
		$curr_rank1++;
		my $next_data1 = $exp_rank1{ $next_exp1 };
		
		if( $next_data1 == $curr_data1 ) {
			$accu_rank1 = $accu_rank1 + $curr_rank1;
			push( @store_exp1, $next_exp1 );
			next;
		}
		
		$tie_rank1 = $accu_rank1 / ( $#store_exp1 + 1 );
		$accu_rank1 = $curr_rank1;
		$curr_data1 = $next_data1;
		
		for( my $i = 0; $i <= $#store_exp1; $i++ ) {
			$exp_rank1{ $store_exp1[$i] } = $tie_rank1;
		}
		@store_exp1 = ();
		push( @store_exp1, $next_exp1 );
	}
	
	$tie_rank1 = $accu_rank1 / ( $#store_exp1 + 1 );
	$accu_rank1 = $curr_rank1;
	
	for( my $i = 0; $i <= $#store_exp1; $i++ ) {
		$exp_rank1{ $store_exp1[$i] } = $tie_rank1;
	}
	
	my $rank_lst = '';
	while( my $exp_id1 = shift( @exp_lst1 ) ) {
		$rank_lst = $rank_lst.$exp_rank1{ $exp_id1 }."\t";
	}
	chop( $rank_lst );
	return $rank_lst;
}
