#!/usr/bin/perl -w
use strict;

my $infile_foldchange = $ARGV[0];
my $infile_geneset = $ARGV[1];
my $infile_ctrl_geneset = $ARGV[2];
my $weight = $ARGV[3]; # 0 or 1
my $type = $ARGV[4]; # "abs" sorts accodring to absolute value
                     # "ori" sorts according to original value
                     # "rev" sorts according to minus(original) value
my $outfile1 = $ARGV[5];

# assign the median of the fold changes as the differential scores in GSEA
open( in_foldchange, "<$infile_foldchange" );
my $header1 = <in_foldchange>;
my %fc_median = ();
while( my $line1 = <in_foldchange> ) {
	chomp( $line1 );
	my ( $id1, @data1 ) = split( /\t/, $line1 );
	my $median1 = median( @data1 );
	
	if( $type eq 'abs' ) { $median1 = abs( $median1 ) };
	if( $type eq 'rev' ) { $median1 = - $median1 };
	
	if( exists( $fc_median{ $id1 } ) ) {
		print "ERROR 1: Multiple lines for single gene!\n";
		exit;
	}
	
	$fc_median{ $id1 } = $median1;
}
close( in_foldchange );

open( in_geneset, "<$infile_geneset" );
my %geneset_gsea = ();
my %geneset_score = ();
my %geneset_p = ();
while( my $line2 = <in_geneset> ) {
	chomp( $line2 );
	my ( $geneset1, $size1, @gene_score1 ) = split( /\t/, $line2 );
	my @gene1 = ();
	
	for( my $i = 0; $i <= $#gene_score1; $i++ ) {
		my ( $gene, $score ) = split( /:/, $gene_score1[$i] );
		push( @gene1, $gene );
	}
	
	my $gsea1 = gsea( \@gene1, \%fc_median, $weight, "all" );
	my @arr1 = split( /\t/, $gsea1, 2 );
	$geneset_score{ $geneset1 } = $arr1[0];
	$geneset_gsea{ $geneset1 } = $gsea1;
	$geneset_p{ $geneset1 } = 0;
}
close( in_geneset );

my $n_ctrl = 0;
open( in_ctrl_geneset, "<$infile_ctrl_geneset" );
while( my $line3 = <in_ctrl_geneset> ) {
	chomp( $line3 );
	
	if( $line3 =~m/^>/ ) {
		my @arr1 = split( /\s/, $line3 );
		$n_ctrl = pop( @arr1 );
		chop( $n_ctrl );
		
		print "Processing Ctrl GeneSet $n_ctrl...\n";
		next;
	}
	
	if( $line3 =~m/^\// ) {
		next;
	}
	
	my ( $geneset1, $size1, @ctrl_gene_score1 ) = split( /\t/, $line3 );
	my @ctrl_gene1 = ();
	for( my $i = 0; $i <= $#ctrl_gene_score1; $i++ ) {
		my ( $gene, $score ) = split( /:/, $ctrl_gene_score1[$i] );
		push( @ctrl_gene1, $gene );
	}
	
	my $ctrl_gsea_score = gsea( \@ctrl_gene1, \%fc_median, $weight, "only_score" );
	if( $ctrl_gsea_score >= $geneset_score{ $geneset1 } ) {
		$geneset_p{ $geneset1 }++;
	}
}
close( in_ctrl_geneset );


open( out1, ">$outfile1" );
my @output1 = sort { $geneset_p{ $a } <=> $geneset_p{ $b } } keys %geneset_p;
for( my $i = 0; $i <= $#output1; $i++ ) {
	my $key1 = $output1[$i];
	print out1 join( "\t", $key1, $geneset_p{ $key1 } / $n_ctrl, $geneset_gsea{ $key1 } );
	print out1 "\n";
}
close( out1 );

###################################################
sub median {
	my @data=sort {$a <=> $b} @_;
	if(int($#data/2)==$#data/2) {
		return $data[$#data/2];
	}
	else {
		return ($data[int($#data/2)]+$data[int($#data/2)+1])/2;
	}
}

# gsea
sub gsea {
	my ( $array_ref, $hash_all_score, $score_weight, $return_type ) = @_;
	
	# մӴС˳
	my @sort_all = sort{ $hash_all_score->{ $b } <=> $hash_all_score->{ $a } } keys %$hash_all_score;
	my %hash_ref = ();
	
	# ο򼯵ܵ÷
	my $ref_score_tot = 0;
	# 㱳ϵĴС
	my $background_n = 0;
	# ӵscoreݼȨȨȷ0Ϊ11Ϊabs(input_score)
	my $add_score = 0;
	
	for( my $i = 0; $i < scalar( @$array_ref ); $i++ ) {
		if( exists( $hash_all_score->{ $array_ref->[$i] } ) ) {
			if( $score_weight == 0 ) {
				$add_score = 1;
			}
			else {
				$add_score = abs( $hash_all_score->{ $array_ref->[$i] } );
			}
			
			$ref_score_tot = $ref_score_tot + $add_score;
			$background_n++;
		}
		
		$hash_ref{ $array_ref->[$i] } = "yes";
	}
	$background_n = $#sort_all + 1 - $background_n;
	#exit;
	
	# gsea
	my $max_up = 0; # ֵ
	my $lead_up_i = -1; # ֵӦбλ
	my $cscore = 0; # ǰķֵ
	for( my $i = 0; $i <= $#sort_all; $i++ ) {
		my $cid1 = $sort_all[$i];
		#print join( "\t", $cid1, $hash_all_score->{ $cid1 } );
		#print "\n";
		
		#ڻ򼯺Уhitмӷ֣missмӷ
		if( exists( $hash_ref{ $cid1 } ) ) {
			if( $score_weight == 0 ) {
				$add_score = 1;
			}
			else {
				$add_score = abs( $hash_all_score->{ $cid1 } );
			}
			
			$cscore = $cscore + $add_score / $ref_score_tot;
		}
		else {
			$cscore = $cscore - 1 / $background_n;
		}
		
		if( $cscore > $max_up ) {
			$max_up = $cscore;
			$lead_up_i = $i;
		}
	}
	
	if( $max_up == 0 ) {
		if( $return_type ne 'only_score' ) {
			return join( "\t", $max_up, "-NA-" );
		}
		else {
			return $max_up;
		}
	}
	
	my @lead_edge = ();
	for( my $i = 0; $i <= $lead_up_i; $i++ ) {
		my $cid1 = $sort_all[$i];
		
		if( exists( $hash_ref{ $cid1 } ) ) {
			push( @lead_edge, $cid1 );
		}
	}
	
	if( $return_type ne 'only_score' ) {
		return join( "\t", $max_up, @lead_edge );
	}
	else {
		return $max_up;
	}
}

sub ref_rand {
	my ( $all_gene_lst, $rand_lst, $n_rand ) = @_;
	my @all_gene = @$all_gene_lst;
	
	for( my $i = 0; $i < $n_rand; $i++ ) {
		my $j = int( rand( $#all_gene + 1 ) );
		$rand_lst->[$i] = splice( @all_gene, $j, 1 );
	}
}
