#!/usr/bin/perl -w
use strict;

print "*** PROGRAM starts: ";
print scalar localtime(time()), " ***\n";

my $infile1_geneset = $ARGV[0];
my $infile2_ctrl = $ARGV[1];
my $infile3_cluster = $ARGV[2];
### my $outfile1 = $ARGV[3];

my $min_overlap_n = 2;

# Reading clusters
# Cluster_ID<\t>Cluster_GeneList
open( in_cluster, "<$infile3_cluster" );
my @cluster1 = <in_cluster>;
chomp( @cluster1 );
close( in_cluster );

# calculate the summarized targetscan scores for all pairs of miRNA-cluster in all background gene sets
print "Calculate the summarized targetscan scores per miRNA per cluster across all runs.\n";
my %ctrl_mirfam_cluster_score = (); # record the scores for each pair of miRNA-cluster, separated by "\t" (N=1000/10000)
my @arr_mirfam_cluster = (); # record the list of miRNA-cluster pairs. Each pair will have N (1000/10000) random scores
my $i_run = 0; # record the RUN
open( in_ctrl, "<$infile2_ctrl" );
while( my $line1 = <in_ctrl> ) {
	# The index of current RUN
	if( $line1 =~ m/^>/ ) {
		$i_run++;
		###print "Reading RUN $i_run....\n";
		next;
	}
	
	if( $line1 =~ m/^\// ) {
		###if( $i_run == 100 ) { last; }
		next;
	}
	
	chomp( $line1 );
	# miRFam, number of target, the target gene list separated by <\tab>
	my ( $mirfam1, $target_size1, @target1 ) = split( /\t/, $line1 );
	
	# Calculate the summarized scores between each cluster and each miRNA
	for( my $i = 0; $i <= $#cluster1; $i++ ) {
		my ( $cid1, $cluster_size1, $cluster_tag1, @cluster_gene1 ) = split( /\t/, $cluster1[$i] );
		# Calculate the meet score using sub-routine meet_score()
		my ( $mirfam_cluster_score1, $overlap_n, $overlap1 ) = split( /\t/, meet_score( \@target1, \@cluster_gene1 ) );
		
		my $mirfam_cluster = "$mirfam1\t$cid1";
		
		if( !exists( $ctrl_mirfam_cluster_score{ $mirfam_cluster } ) ) {
			$ctrl_mirfam_cluster_score{ $mirfam_cluster } = $mirfam_cluster_score1;
			push( @arr_mirfam_cluster, $mirfam_cluster );
		}
		else {
			$ctrl_mirfam_cluster_score{ $mirfam_cluster } = $ctrl_mirfam_cluster_score{ $mirfam_cluster }."\t".$mirfam_cluster_score1;
		}
	}
}
close( in_ctrl );

# calculate rank for each pair of miRNA-cluster using ctrl gene sets
print "Calculate the rank per miRNA per cluster across all runs.\n";
my %ctrl_run_mirfam_cluster_pscore = (); # The ranks (p-value = rank/N) according to the miRFam:Cluster summarized scores
                                         # hash key: i_Run<\t>miRFam<\t>Cluster
for( my $i = 0; $i <= $#arr_mirfam_cluster; $i++ ) {
	my $mirfam_cluster = $arr_mirfam_cluster[$i];
	
	# Calculate the ranks (p-value = rank/N) according to the miRFam:Cluster summarized scores
	#   the miRFam:Cluster pair with bigger score will rank on top (with small rank or p-value)
	my @rank1 = split( /\t/, get_low_rank( $ctrl_mirfam_cluster_score{ $mirfam_cluster }, "d" ) );

	for( my $j = 0; $j <= $#rank1; $j++ ) {
		my $pscore1 = $rank1[$j] - 1;
		#  i_Run --- $j + 1
		#  miRFam:Cluster --- $mirfam_cluster
		my $run_mirfam_cluster = join( "\t", $j + 1, $mirfam_cluster );
		
		# Record the rank (or p-value) for each i_Run:miRFam:Cluster
		$ctrl_run_mirfam_cluster_pscore{ $run_mirfam_cluster } = $rank1[$j];
	}	
}

# Calculate the minimal rank for each miRam across all clusters
#  hash key --- i_Run:miRFam
print "Calculate the minimal rank for per miRNA per run across all clusters.\n";
my %ctrl_run_mirfam_pscore = (); # The hash for the minimal rank for each i_Run:miRFam across all clusters
                                 #   eliminate the variable Cluster by min(Cluster) [i_Run:miRFam:Cluster]
                                 #   equal to the final p-value of each control gene set
my @arr_mirfam = (); # The array for miRFam
my %hash_mirfam = (); # The hash for miRFam

for( my $i = 0; $i <= $#arr_mirfam_cluster; $i++ ) { # miRFam:Cluster
	my $mirfam_cluster = $arr_mirfam_cluster[$i];
	my ( $mirfam, $cluster ) = split( /\t/, $mirfam_cluster );
	
	# Record all miRFams
	if( !exists( $hash_mirfam{ $mirfam } ) ) {
		$hash_mirfam{ $mirfam } = "yes";
		push( @arr_mirfam, $mirfam );
	}
	
	for( my $j = 0; $j < $i_run; $j++ ) { # i_Run in miRFam:Cluster
		my $run_mirfam_cluster = join( "\t", $j + 1, $mirfam_cluster );
		my $run_mirfam = join( "\t", $j + 1, $mirfam );
		
		if( !exists( $ctrl_run_mirfam_pscore{ $run_mirfam } ) ) { # Record the first rank for i_Run:miRFam
			$ctrl_run_mirfam_pscore{ $run_mirfam } = $ctrl_run_mirfam_cluster_pscore{ $run_mirfam_cluster };
		}
		else {
			# Record the minimal rank for i_Run:miRFam
			if( $ctrl_run_mirfam_cluster_pscore{ $run_mirfam_cluster } < $ctrl_run_mirfam_pscore{ $run_mirfam } ) {
				$ctrl_run_mirfam_pscore{ $run_mirfam } = $ctrl_run_mirfam_cluster_pscore{ $run_mirfam_cluster };
			}
		}
	}
}

# Transform run_mirfam_pscore to mirfam_pscore list (separated by "\t")
print "\n*** P-score for each ctrl miRNA set in each RUN. ***\n";
my %ctrl_mirfam_pscore = (); # The pscore/p-value for each miRFam
for( my $i = 0; $i <= $#arr_mirfam; $i++ ) { # For each miRFam
	my $mirfam = $arr_mirfam[$i];
	
	my $mirfam_pscore = "";
	for( my $j = 0; $j < $i_run; $j++ ) {
		my $run_mirfam = join( "\t", $j + 1, $mirfam );
		
		if( $j == 0 ) {
			$mirfam_pscore = $ctrl_run_mirfam_pscore{ $run_mirfam };
		}
		else {
			$mirfam_pscore = $mirfam_pscore."\t".$ctrl_run_mirfam_pscore{ $run_mirfam };
		}
	}
	
	$ctrl_mirfam_pscore{ $mirfam } = $mirfam_pscore;
	
	print join( "\t", $mirfam, $mirfam_pscore ), "\n";
}
print "*** P-score for each ctrl miRNA set in each RUN. ***\n\n";

############## exit;
# ------------------------------------------------ #
my %mirfam_cluster_score_info = ();
my %mirfam_cluster_pscore = ();
my %mirfam_pscore_info = ();
my %mirfam_pscore = ();

open( in_geneset, "<$infile1_geneset" );
print "\n*** P-score for each miRNA target set in each cluster. ***\n";
while( my $line1 = <in_geneset> ) {
	chomp( $line1 );
	my ( $mirfam1, $target_size1, @target1 ) = split( /\t/, $line1 );
	
	print $mirfam1;
	for( my $i = 0; $i <= $#cluster1; $i++ ) { # miRFam along each Cluster
		my ( $cid1, $cluster_size1, $cluster_tag1, @cluster_gene1 ) = split( /\t/, $cluster1[$i] );
		my ( $mirfam_cluster_score1, $overlap_n, $overlap1 ) = split( /\t/, meet_score( \@target1, \@cluster_gene1 ) );
		
		my $mirfam_cluster = "$mirfam1\t$cid1";
		
		# calculate the pscore/rank 
		my @ctrl_score = split( /\t/, $ctrl_mirfam_cluster_score{ $mirfam_cluster } );
		my $pscore1 = 0;
		
		if( $#ctrl_score + 1 != $i_run ) {
			print "Checking error!! Error Code 1.\n";
			exit;
		}
		
		# Calculate the p-value for each miRFam:Cluster
		for( my $i = 0; $i <= $#ctrl_score; $i++ ) {
			if( $ctrl_score[$i] >= $mirfam_cluster_score1 ) {
				$pscore1 ++;
			}
		}
		
		$mirfam_cluster_pscore{ $mirfam_cluster } = $pscore1;
		$mirfam_cluster_score_info{ $mirfam_cluster } = join( "\t", $mirfam_cluster_score1, $pscore1, $overlap_n, $target_size1,
		                                                            $cid1, $cluster_size1, $overlap1 );
		
		print "\t".$pscore1;
		if( !exists( $mirfam_pscore{ $mirfam1 } ) ) {
			$mirfam_pscore{ $mirfam1 } = $pscore1;
			$mirfam_pscore_info{ $mirfam1 } = $mirfam_cluster_score_info{ $mirfam_cluster };
		}
		else {
			if( $pscore1 < $mirfam_pscore{ $mirfam1 } ) {
				$mirfam_pscore{ $mirfam1 } = $pscore1;
				$mirfam_pscore_info{ $mirfam1 } = $mirfam_cluster_score_info{ $mirfam_cluster };
			}
		}
	}
	
	print "\n";
}
close( in_geneset );
print "*** P-score for each miRNA target set in each cluster. ***\n\n";

print "-------------------------------------\n";
print "---------- miRHiC results. ----------\n\n";
my %mirfam_pvalue = ();
for( my $i = 0; $i <= $#arr_mirfam; $i++ ) {
	my $mirfam1 = $arr_mirfam[$i];
	my $pscore1 = $mirfam_pscore{ $mirfam1 };
	my @ctrl_score = split( /\t/, $ctrl_mirfam_pscore{ $mirfam1 } );
	
	# calculate the final p-value
	my $pvalue = 0;
	for( my $i = 0; $i <= $#ctrl_score; $i++ ) {
		my $ctrl_score1 = $ctrl_score[$i];
		
		if( $ctrl_score1 <= $pscore1 ) {
			$pvalue++;
		}
	}
	
	$pvalue = $pvalue / ( $#ctrl_score + 1 );
	print join( "\t", $mirfam1, $pvalue, $mirfam_pscore_info{ $mirfam1 } ), "\n";
}

print "*** PROGRAM stops: ";
print scalar localtime(time()), " ***\n";

# The sub-routine to calculate the summarized score of the overlapped genes between 
#    the miRNA target gene set (with weights or scores) and the cluster gene set
sub meet_score {
	my ( $arr_target_score, $arr_cluster ) = @_;
	my @cluster_gene = @$arr_cluster;
	
	my %hash_gene = (); # record all genes in the cluster
	for( my $i = 0; $i <= $#cluster_gene; $i++ ) {
		$hash_gene{ $cluster_gene[$i] } = "yes";
	}
	
	my @target_score = @$arr_target_score;
	my $sum_score = 0; # the summarized score
	my $n_overlap = 0; # the number of the overlapped genes
	my $overlap = ''; # the list of the overlapped genes
	for( my $i = 0; $i <= $#target_score; $i++ ) {
		my ( $target1, $score1 ) = split( /:/, $target_score[$i] );
		
		if( exists( $hash_gene{ $target1 } ) ) { # if this target gene is in the cluster gene list, add its score
			$sum_score = $sum_score + $score1;
			$n_overlap++;
			if( $n_overlap == 1 ) {
				$overlap = $target_score[$i];
			}
			else {
				$overlap = $overlap."|".$target_score[$i];
			}
		}
	}
	
	return join( "\t", $sum_score, $n_overlap, $overlap );
}

# Calculate the rank list of input data. If tie, return the lowest rank.
sub get_low_rank {
	my ( $data_lst1, $order1 ) = @_; # $order = "i"/"d"
	my @data_exp1 = split( /\t/, $data_lst1 ); # the data array
		
	my %exp_rank1 = (); # the data hash with names
	my @exp_lst1 = (); # the names of above hash
	
	for( my $i = 0; $i <= $#data_exp1; $i++ ) {
		$exp_rank1{ "data$i" } = $data_exp1[$i];
		push( @exp_lst1, "data$i" );
	}
	
	# get the names according to the sorted data
	my @sort_exp_rank1 = ();
	if( $order1 eq 'i' ) {
		@sort_exp_rank1 = sort { $exp_rank1{ $a } <=> $exp_rank1{ $b } } keys %exp_rank1;
	}
	else {
		@sort_exp_rank1 = sort { $exp_rank1{ $b } <=> $exp_rank1{ $a } } keys %exp_rank1;
	}
	
	my $curr_data1 = $exp_rank1{ $sort_exp_rank1[0] };
	my $accu_rank1 = 1;
	my $curr_rank1 = 1;
	my @store_exp1 = ();
	my $tie_rank1 = 0;
	
	push( @store_exp1, shift( @sort_exp_rank1 ) );
	
	while( my $next_exp1 = shift( @sort_exp_rank1 ) ) {
		$curr_rank1++;
		my $next_data1 = $exp_rank1{ $next_exp1 };
		
		if( $next_data1 == $curr_data1 ) {
			#$accu_rank1 = $accu_rank1 + $curr_rank1;
			$accu_rank1 = $curr_rank1; # Record the "worst" rank for calculating p-values
			push( @store_exp1, $next_exp1 );
			next;
		}
		
		#$tie_rank1 = $accu_rank1 / ( $#store_exp1 + 1 );
		$tie_rank1 = $accu_rank1;
		$accu_rank1 = $curr_rank1;
		$curr_data1 = $next_data1;
		
		for( my $i = 0; $i <= $#store_exp1; $i++ ) {
			$exp_rank1{ $store_exp1[$i] } = $tie_rank1;
		}
		
		@store_exp1 = ();
		push( @store_exp1, $next_exp1 );
	}
	
	# $tie_rank1 = $accu_rank1 / ( $#store_exp1 + 1 );
	$tie_rank1 = $accu_rank1;
	$accu_rank1 = $curr_rank1;
	
	for( my $i = 0; $i <= $#store_exp1; $i++ ) {
		$exp_rank1{ $store_exp1[$i] } = $tie_rank1;
	}
	
	my $rank_lst = '';
	while( my $exp_id1 = shift( @exp_lst1 ) ) {
		$rank_lst = $rank_lst.$exp_rank1{ $exp_id1 }."\t";
	}
	chop( $rank_lst );
	return $rank_lst;
}
